import os
import json
import asyncio
import inspect
import functools
import random

import uvicorn
from typing_extensions import Annotated
from fastapi import FastAPI, Request, Security, status, HTTPException
from fastapi.security import APIKeyHeader, APIKeyQuery
from fastapi.responses import StreamingResponse
from pydantic import BaseModel, ConfigDict, Field, BeforeValidator, ValidationError

from . import logger
from .utils import convert_to_bytes, generate_data

class MaxSizePerRequestError(Exception):
    pass


class MinSizePerRequestError(Exception):
    pass


class Testdata:

    class Config(BaseModel):

        model_config = ConfigDict(extra='forbid')

        @staticmethod
        def to_bytes(value: int | str) -> int:
            try:
                return convert_to_bytes(value)
            except Exception as err:
                raise ValidationError from err

        @staticmethod
        def is_authorized_keys(value: set[str] | str) -> set[str]:
            if isinstance(value, str):
                with open(value, encoding='utf-8') as file:
                    return set(filter(lambda x: x.strip() != '', file.read().splitlines()))
            return value

        authorized_keys: Annotated[set[str], BeforeValidator(is_authorized_keys)] = Field(alias='keys')
        max_size: Annotated[int, BeforeValidator(to_bytes)] = Field(alias='max-size')
        max_data: Annotated[int, BeforeValidator(to_bytes)] = Field(alias='max-data')
        buffer_size: Annotated[int, BeforeValidator(to_bytes)] = Field(alias='buffer-size')
        database: str | None = None
        log: str | None = Field(alias='log', default=None)
        database_update_interval: float = Field(alias='database-update-interval', default=5)

    _config: Config
    _api: FastAPI
    _state: dict[str, int]
    _logger: logger.Logger

    def __init__(self, config: Config):
        self._config = config
        self._logger = logger.getLogger('testdata')
        self._api = self._setup_api()

        # Store internal state
        self._state = {'data-used': 0}

    def _setup_api(self) -> FastAPI:
        api = FastAPI(docs_url='/', redoc_url=None)

        # Security
        def get_api_key(
            api_key_query: str = Security(APIKeyQuery(name="api_key", auto_error=False)),
            api_key_header: str = Security(APIKeyHeader(name="x-api-key", auto_error=False))
        ) -> str:
            # https://joshdimella.com/blog/adding-api-key-auth-to-fast-api

            if api_key_query in self._config.authorized_keys:
                return api_key_query
            if api_key_header in self._config.authorized_keys:
                return api_key_header

            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail='Invalid or missing API Key'
            )

        # A wrapper to set the function signature to accept the api key dependency
        def secure(func):
            # Get old signature
            positional_only, positional_or_keyword, variadic_positional, keyword_only, variadic_keyword = [], [], [], [], []
            for value in inspect.signature(func).parameters.values():
                if value.kind == inspect.Parameter.POSITIONAL_ONLY:
                    positional_only.append(value)
                elif value.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
                    positional_or_keyword.append(value)
                elif value.kind == inspect.Parameter.VAR_POSITIONAL:
                    variadic_positional.append(value)
                elif value.kind == inspect.Parameter.KEYWORD_ONLY:
                    keyword_only.append(value)
                elif value.kind == inspect.Parameter.VAR_KEYWORD:
                    variadic_keyword.append(value)

            # Avoid passing an unrecognized keyword
            if inspect.iscoroutinefunction(func):
                async def wrapper(*args, **kwargs):
                    if len(variadic_keyword) == 0:
                        if 'api_key' in kwargs:
                            del kwargs['api_key']

                    return await func(*args, **kwargs)
            else:
                def wrapper(*args, **kwargs):
                    if len(variadic_keyword) == 0:
                        if 'api_key' in kwargs:
                            del kwargs['api_key']

                    return func(*args, **kwargs)

            # Override signature
            wrapper.__signature__ = inspect.signature(func).replace(
                parameters=(
                    *positional_only,
                    *positional_or_keyword,
                    *variadic_positional,
                    *keyword_only,
                    inspect.Parameter('api_key', inspect.Parameter.POSITIONAL_OR_KEYWORD, default=Security(get_api_key)),
                    *variadic_keyword
                )
            )

            return functools.wraps(func)(wrapper)

        # Routes
        api.get('/zeros')(secure(self._zeros))

        return api

    async def _zeros(self, size: int | str, request: Request, filename: str = 'zeros.bin') -> StreamingResponse:
        try:
            extra = {'id': f'{random.randint(0, 2 ** 32 - 1):08X}'}
            self._logger.debug(
                'Initiated request.',
                extra=extra | {
                    'ip': request.client.host if request.client is not None else None,
                    'query-params': dict(request.query_params),
                    'headers': dict(request.headers)
                }
            )

            try:
                size = convert_to_bytes(size)
            except ValueError as err:
                self._logger.warning('Invalid format for size.', extra=extra)
                raise HTTPException(
                    status_code=status.HTTP_400_BAD_REQUEST,
                    detail='Invalid format for size.'
                ) from err

            if size < 0:
                raise MinSizePerRequestError
            if self._config.max_size < size:
                raise MaxSizePerRequestError

            # update internal state
            if self._config.max_data < self._state['data-used'] + size:
                raise HTTPException(
                    status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                    detail='Service not available.'
                )
            self._state['data-used'] += size

            self._logger.debug('Successfully processed request.', extra=extra)
            return StreamingResponse(
                status_code=status.HTTP_200_OK,
                content=generate_data(size, self._config.buffer_size),
                media_type='application/octet-stream',
                headers={
                    'Content-Length': str(size),
                    'Content-Disposition': f'attachment; filename="{filename}"'
                }
            )

        except MinSizePerRequestError as err:
            self._logger.warning('Size if negative.', extra=extra)
            raise HTTPException(
                status_code=status.HTTP_416_REQUESTED_RANGE_NOT_SATISFIABLE,
                detail='Size has to be non-negative.'
            ) from err
        except MaxSizePerRequestError as err:
            self._logger.warning('Exceeded max size per request.', extra=extra)
            raise HTTPException(
                status_code=status.HTTP_416_REQUESTED_RANGE_NOT_SATISFIABLE,
                detail=f'Exceeded max size per request of {self._config.max_size} Bytes.'
            ) from err
        except Exception as err:
            self._logger.exception(err)
            raise err

    async def _update_state(self) -> None:
        assert self._config.database is not None

        mode = 'r+' if os.path.exists(self._config.database) else 'w+'

        with open(self._config.database, mode, encoding='utf-8') as file:
            try:
                self._state = json.load(file)
            except json.JSONDecodeError:
                pass

            while True:
                file.seek(0)
                json.dump(self._state, file)
                file.truncate()
                await asyncio.sleep(self._config.database_update_interval)

    async def run(self, host: str, port: int) -> None:
        try:
            if self._config.log is not None:
                logger.setup_logging(self._config.log)

            self._logger = logger.getLogger('testdata')
            self._logger.info('Server started.')

            coroutines = [uvicorn.Server(uvicorn.Config(self._api, host, port)).serve()]
            if self._config.database is not None:
                coroutines.append(self._update_state())

            await asyncio.gather(*coroutines)
        except asyncio.exceptions.CancelledError:
            self._logger.info('Server stopped.')
        except Exception as err:
            self._logger.exception(err)
